/*
 * Copyright (c) 1995-2002, Index Data
 * See the file LICENSE for details.
 *
 * $Id: client.c,v 1.147 2002/03/20 14:35:42 adam Exp $
 */

#include <stdio.h>
#include <stdlib.h>
#include <time.h>

#include <yaz/yaz-util.h>

#include <yaz/tcpip.h>
#ifdef USE_XTIMOSI
#include <yaz/xmosi.h>
#endif

#include <yaz/proto.h>
#include <yaz/marcdisp.h>
#include <yaz/diagbib1.h>
#include <yaz/otherinfo.h>

#include <yaz/pquery.h>
#include <yaz/sortspec.h>

#if YAZ_MODULE_ill
#include <yaz/ill.h>
#endif

#if YAZ_MODULE_ccl
#include <yaz/yaz-ccl.h>
#endif

#if HAVE_READLINE_READLINE_H
#include <readline/readline.h>
#endif
#if HAVE_READLINE_HISTORY_H
#include <readline/history.h>
#endif

#include <sys/stat.h>


#include "admin.h"
#include "tabcomplete.h"

#define C_PROMPT "Z> "

static ODR out, in, print;              /* encoding and decoding streams */
static FILE *apdu_file = 0;
static COMSTACK conn = 0;               /* our z-association */
static Z_IdAuthentication *auth = 0;    /* our current auth definition */
char *databaseNames[128];
int num_databaseNames = 0;
static Z_External *record_last = 0;
static int setnumber = -1;               /* current result set number */
static int smallSetUpperBound = 0;
static int largeSetLowerBound = 1;
static int mediumSetPresentNumber = 0;
static Z_ElementSetNames *elementSetNames = 0; 
static int setno = 1;                   /* current set offset */
static enum oid_proto protocol = PROTO_Z3950;      /* current app protocol */
static enum oid_value recordsyntax = VAL_USMARC;
static enum oid_value schema = VAL_NONE;
static int sent_close = 0;
static NMEM session_mem = NULL;      /* memory handle for init-response */
static Z_InitResponse *session = 0;     /* session parameters */
static char last_scan_line[512] = "0";
static char last_scan_query[512] = "0";
static char ccl_fields[512] = "default.bib";
static char* esPackageName = 0;
static char* yazProxy = 0;
static int kilobytes = 1024;

static char last_cmd[32] = "?";
static FILE *marcdump = 0;
static char *refid = NULL;

typedef enum {
    QueryType_Prefix,
    QueryType_CCL,
    QueryType_CCL2RPN
} QueryType;

static QueryType queryType = QueryType_Prefix;

#if YAZ_MODULE_ccl
static CCL_bibset bibset;               /* CCL bibset handle */
#endif


/* set this one to 1, to avoid decode of unknown MARCs  */
#define AVOID_MARC_DECODE 1

/* nice helper macro as extensive tabbing gives spaces at the en of the args lines */
#define REMOVE_TAILING_BLANKS(a) {\
  char* args_end=(a)+strlen(a)-1; \
  while(isspace(*args_end)) {*args_end=0;--args_end;}; \
  }


void process_cmd_line(char* line);
char ** readline_completer(char *text, int start, int end);
char *command_generator(const char *text, int state);


ODR getODROutputStream()
{
    return out;
}

void send_apdu(Z_APDU *a)
{
    char *buf;
    int len;

    if (apdu_file)
    {
        z_APDU(print, &a, 0, 0);
        odr_reset(print);
    }
    if (!z_APDU(out, &a, 0, 0))
    {
        odr_perror(out, "Encoding APDU");
        exit(1);
    }
    buf = odr_getbuf(out, &len, 0);
    /* printf ("sending APDU of size %d\n", len); */
    if (cs_put(conn, buf, len) < 0)
    {
        fprintf(stderr, "cs_put: %s", cs_errmsg(cs_errno(conn)));
        exit(1);
    }
    odr_reset(out); /* release the APDU structure  */
}

static void print_stringn(const unsigned char *buf, size_t len)
{
   size_t i;
   for (i = 0; i<len; i++)
       if ((buf[i] <= 126 && buf[i] >= 32) || strchr ("\n\r\t\f", buf[i]))
           printf ("%c", buf[i]);
       else
           printf ("\\X%02X", buf[i]);
}

static void print_refid (Z_ReferenceId *id)
{
    if (id)
    {
        printf ("Reference Id: ");
        print_stringn (id->buf, id->len);
        printf ("\n");
    }
}

static Z_ReferenceId *set_refid (ODR out)
{
    Z_ReferenceId *id;
    if (!refid)
        return 0;
    id = (Z_ReferenceId *) odr_malloc (out, sizeof(*id));
    id->size = id->len = strlen(refid);
    id->buf = (unsigned char *) odr_malloc (out, id->len);
    memcpy (id->buf, refid, id->len);
    return id;
}   

/* INIT SERVICE ------------------------------- */

static void send_initRequest(const char* type_and_host)
{
    Z_APDU *apdu = zget_APDU(out, Z_APDU_initRequest);
    Z_InitRequest *req = apdu->u.initRequest;

    ODR_MASK_SET(req->options, Z_Options_search);
    ODR_MASK_SET(req->options, Z_Options_present);
    ODR_MASK_SET(req->options, Z_Options_namedResultSets);
    ODR_MASK_SET(req->options, Z_Options_triggerResourceCtrl);
    ODR_MASK_SET(req->options, Z_Options_scan);
    ODR_MASK_SET(req->options, Z_Options_sort);
    ODR_MASK_SET(req->options, Z_Options_extendedServices);
    ODR_MASK_SET(req->options, Z_Options_delSet);

    ODR_MASK_SET(req->protocolVersion, Z_ProtocolVersion_1);
    ODR_MASK_SET(req->protocolVersion, Z_ProtocolVersion_2);
    ODR_MASK_SET(req->protocolVersion, Z_ProtocolVersion_3);

    *req->maximumRecordSize = 1024*kilobytes;
    *req->preferredMessageSize = 1024*kilobytes;

    req->idAuthentication = auth;

    req->referenceId = set_refid (out);

    if (yazProxy) 
        yaz_oi_set_string_oidval(&req->otherInfo, out, VAL_PROXY,
        1, type_and_host);
    
    send_apdu(apdu);
    printf("Sent initrequest.\n");
}

static int process_initResponse(Z_InitResponse *res)
{
    /* save session parameters for later use */
    session_mem = odr_extract_mem(in);
    session = res;

    if (!*res->result)
        printf("Connection rejected by target.\n");
    else
        printf("Connection accepted by target.\n");
    if (res->implementationId)
        printf("ID     : %s\n", res->implementationId);
    if (res->implementationName)
        printf("Name   : %s\n", res->implementationName);
    if (res->implementationVersion)
        printf("Version: %s\n", res->implementationVersion);
    if (res->userInformationField)
    {
        printf("UserInformationfield:\n");
        if (!z_External(print, (Z_External**)&res-> userInformationField,
            0, 0))
        {
            odr_perror(print, "Printing userinfo\n");
            odr_reset(print);
        }
        if (res->userInformationField->which == Z_External_octet)
        {
            printf("Guessing visiblestring:\n");
            printf("'%s'\n", res->userInformationField->u. octet_aligned->buf);
        }
        odr_reset (print);
    }
    printf ("Options:");
    if (ODR_MASK_GET(res->options, Z_Options_search))
        printf (" search");
    if (ODR_MASK_GET(res->options, Z_Options_present))
        printf (" present");
    if (ODR_MASK_GET(res->options, Z_Options_delSet))
        printf (" delSet");
    if (ODR_MASK_GET(res->options, Z_Options_resourceReport))
        printf (" resourceReport");
    if (ODR_MASK_GET(res->options, Z_Options_resourceCtrl))
        printf (" resourceCtrl");
    if (ODR_MASK_GET(res->options, Z_Options_accessCtrl))
        printf (" accessCtrl");
    if (ODR_MASK_GET(res->options, Z_Options_scan))
        printf (" scan");
    if (ODR_MASK_GET(res->options, Z_Options_sort))
        printf (" sort");
    if (ODR_MASK_GET(res->options, Z_Options_extendedServices))
        printf (" extendedServices");
    if (ODR_MASK_GET(res->options, Z_Options_level_1Segmentation))
        printf (" level1Segmentation");
    if (ODR_MASK_GET(res->options, Z_Options_level_2Segmentation))
        printf (" level2Segmentation");
    if (ODR_MASK_GET(res->options, Z_Options_concurrentOperations))
        printf (" concurrentOperations");
    if (ODR_MASK_GET(res->options, Z_Options_namedResultSets))
    {
        printf (" namedResultSets");
        setnumber = 0;
    }
    if (ODR_MASK_GET(res->options, Z_Options_encapsulation))
        printf (" encapsulation");
    if (ODR_MASK_GET(res->options, Z_Options_resultCount))
        printf (" resultCount");
    if (ODR_MASK_GET(res->options, Z_Options_negotiationModel))
        printf (" negotiationModel");
    if (ODR_MASK_GET(res->options, Z_Options_duplicateDetection))
        printf (" duplicateDetection");
    if (ODR_MASK_GET(res->options, Z_Options_queryType104))
        printf (" queryType104");
    printf ("\n");
    fflush (stdout);
    return 0;
}

static int cmd_base(char *arg)
{
    int i;
    char *cp;

    if (!*arg)
    {
        printf("Usage: base <database> <database> ...\n");
        return 0;
    }
    for (i = 0; i<num_databaseNames; i++)
        xfree (databaseNames[i]);
    num_databaseNames = 0;
    while (1)
    {
        if (!(cp = strchr(arg, ' ')))
            cp = arg + strlen(arg);
        if (cp - arg < 1)
            break;
        databaseNames[num_databaseNames] = (char *)xmalloc (1 + cp - arg);
        memcpy (databaseNames[num_databaseNames], arg, cp - arg);
        databaseNames[num_databaseNames++][cp - arg] = '\0';
        if (!*cp)
            break;
        arg = cp+1;
    }
    return 1;
}

int cmd_open(char *arg)
{
    void *add;
    char type_and_host[101], base[101];
    CS_TYPE t;
    
    if (conn)
    {
        printf("Already connected.\n");
        
        cs_close (conn);
        conn = NULL;
        if (session_mem)
        {
            nmem_destroy (session_mem);
            session_mem = NULL;
        }
    }
    t = tcpip_type;
    base[0] = '\0';
    if (sscanf (arg, "%100[^/]/%100s", type_and_host, base) < 1)
        return 0;
    
    if(yazProxy) 
        conn = cs_create_host(yazProxy, 1, &add);
    else 
        conn = cs_create_host(type_and_host, 1, &add);
    
    if (!conn)
    {
        printf ("Couldn't create comstack\n");
        return 0;
    }
    printf("Connecting...");
    fflush(stdout);
    if (cs_connect(conn, add) < 0)
    {
        printf ("error = %s\n", cs_strerror(conn));
        if (conn->cerrno == CSYSERR)
            perror("system");
        cs_close(conn);
        conn = 0;
        return 0;
    }
    printf("Ok.\n");
    
    send_initRequest(type_and_host);
    if (*base)
        cmd_base (base);
    return 2;
}

int cmd_authentication(char *arg)
{
    static Z_IdAuthentication au;
    static char user[40], group[40], pass[40];
    static Z_IdPass idPass;
    int r;

    if (!*arg)
    {
        printf("Auth field set to null\n");
        auth = 0;
        return 1;
    }
    r = sscanf (arg, "%39s %39s %39s", user, group, pass);
    if (r == 0)
    {
        printf("Auth field set to null\n");
        auth = 0;
    }
    if (r == 1)
    {
        auth = &au;
        au.which = Z_IdAuthentication_open;
        au.u.open = user;
    }
    if (r == 3)
    {
        auth = &au;
        au.which = Z_IdAuthentication_idPass;
        au.u.idPass = &idPass;
        idPass.groupId = group;
        idPass.userId = user;
        idPass.password = pass;
    }
    return 1;
}

/* SEARCH SERVICE ------------------------------ */

static void display_variant(Z_Variant *v, int level)
{
    int i;

    for (i = 0; i < v->num_triples; i++)
    {
        printf("%*sclass=%d,type=%d", level * 4, "", *v->triples[i]->zclass,
            *v->triples[i]->type);
        if (v->triples[i]->which == Z_Triple_internationalString)
            printf(",value=%s\n", v->triples[i]->value.internationalString);
        else
            printf("\n");
    }
}

static void display_grs1(Z_GenericRecord *r, int level)
{
    int i;

    if (!r)
        return;
    for (i = 0; i < r->num_elements; i++)
    {
        Z_TaggedElement *t;

        printf("%*s", level * 4, "");
        t = r->elements[i];
        printf("(");
        if (t->tagType)
            printf("%d,", *t->tagType);
        else
            printf("?,");
        if (t->tagValue->which == Z_StringOrNumeric_numeric)
            printf("%d) ", *t->tagValue->u.numeric);
        else
            printf("%s) ", t->tagValue->u.string);
        if (t->content->which == Z_ElementData_subtree)
        {
            printf("\n");
            display_grs1(t->content->u.subtree, level+1);
        }
        else if (t->content->which == Z_ElementData_string)
            printf("%s\n", t->content->u.string);
        else if (t->content->which == Z_ElementData_numeric)
            printf("%d\n", *t->content->u.numeric);
        else if (t->content->which == Z_ElementData_oid)
        {
            int *ip = t->content->u.oid;
            oident *oent;
            
            if ((oent = oid_getentbyoid(t->content->u.oid)))
                printf("OID: %s\n", oent->desc);
            else
            {
                printf("{");
                while (ip && *ip >= 0)
                    printf(" %d", *(ip++));
                printf(" }\n");
            }
        }
        else if (t->content->which == Z_ElementData_noDataRequested)
            printf("[No data requested]\n");
        else if (t->content->which == Z_ElementData_elementEmpty)
            printf("[Element empty]\n");
        else if (t->content->which == Z_ElementData_elementNotThere)
            printf("[Element not there]\n");
        else
            printf("??????\n");
        if (t->appliedVariant)
            display_variant(t->appliedVariant, level+1);
        if (t->metaData && t->metaData->supportedVariants)
        {
            int c;

            printf("%*s---- variant list\n", (level+1)*4, "");
            for (c = 0; c < t->metaData->num_supportedVariants; c++)
            {
                printf("%*svariant #%d\n", (level+1)*4, "", c);
                display_variant(t->metaData->supportedVariants[c], level + 2);
            }
        }
    }
}


static void print_record(const unsigned char *buf, size_t len)
{
    size_t i = len;
    print_stringn (buf, len);
    /* add newline if not already added ... */
    if (i <= 0 || buf[i-1] != '\n')
        printf ("\n");
}

static void display_record(Z_External *r)
{
    oident *ent = oid_getentbyoid(r->direct_reference);

    record_last = r;
    /*
     * Tell the user what we got.
     */
    if (r->direct_reference)
    {
        printf("Record type: ");
        if (ent)
            printf("%s\n", ent->desc);
        else if (!odr_oid(print, &r->direct_reference, 0, 0))
        {
            odr_perror(print, "print oid");
            odr_reset(print);
        }
    }
    /* Check if this is a known, ASN.1 type tucked away in an octet string */
    if (ent && r->which == Z_External_octet)
    {
        Z_ext_typeent *type = z_ext_getentbyref(ent->value);
        void *rr;

        if (type)
        {
            /*
             * Call the given decoder to process the record.
             */
            odr_setbuf(in, (char*)r->u.octet_aligned->buf,
                r->u.octet_aligned->len, 0);
            if (!(*type->fun)(in, (char **)&rr, 0, 0))
            {
                odr_perror(in, "Decoding constructed record.");
                fprintf(stderr, "[Near %d]\n", odr_offset(in));
                fprintf(stderr, "Packet dump:\n---------\n");
                odr_dumpBER(stderr, (char*)r->u.octet_aligned->buf,
                    r->u.octet_aligned->len);
                fprintf(stderr, "---------\n");
                exit(1);
            }
            /*
             * Note: we throw away the original, BER-encoded record here.
             * Do something else with it if you want to keep it.
             */
            r->u.sutrs = (Z_SUTRS *) rr; /* we don't actually check the type here. */
            r->which = type->what;
        }
    }
    if (ent && ent->value == VAL_SOIF)
        print_record((const unsigned char *) r->u.octet_aligned->buf,
                     r->u.octet_aligned->len);
    else if (r->which == Z_External_octet && r->u.octet_aligned->len)
    {
        const char *octet_buf = (char*)r->u.octet_aligned->buf;
        if (ent->value == VAL_TEXT_XML || ent->value == VAL_APPLICATION_XML ||
            ent->value == VAL_HTML)
        {
            print_record((const unsigned char *) octet_buf,
                         r->u.octet_aligned->len);
        }
        else if (ent->value == VAL_POSTSCRIPT)
        {
            int size = r->u.octet_aligned->len;
            if (size > 100)
                size = 100;
            print_record((const unsigned char *) octet_buf, size);
        }
        else
        {
            if ( 
#if AVOID_MARC_DECODE
                /* primitive check for a marc OID 5.1-29 except 16 */
                ent->oidsuffix[0] == 5 && ent->oidsuffix[1] < 30 &&
                ent->oidsuffix[1] != 16
#else
                1
#endif
                )
            {
                if (marc_display_exl (octet_buf, NULL, 0 /* debug */,
                                      r->u.octet_aligned->len) <= 0)
                {
                    printf ("bad MARC. Dumping as it is:\n");
                    print_record((const unsigned char*) octet_buf,
                                 r->u.octet_aligned->len);
                }
            }
            else
            {
                print_record((const unsigned char*) octet_buf,
                             r->u.octet_aligned->len);
            }
        }
        if (marcdump)
            fwrite (octet_buf, 1, r->u.octet_aligned->len, marcdump);
    }
    else if (ent && ent->value == VAL_SUTRS)
    {
        if (r->which != Z_External_sutrs)
        {
            printf("Expecting single SUTRS type for SUTRS.\n");
            return;
        }
        print_record(r->u.sutrs->buf, r->u.sutrs->len);
    }
    else if (ent && ent->value == VAL_GRS1)
    {
        if (r->which != Z_External_grs1)
        {
            printf("Expecting single GRS type for GRS.\n");
            return;
        }
        display_grs1(r->u.grs1, 0);
    }
    else 
    {
        printf("Unknown record representation.\n");
        if (!z_External(print, &r, 0, 0))
        {
            odr_perror(print, "Printing external");
            odr_reset(print);
        }
    }
}

static void display_diagrecs(Z_DiagRec **pp, int num)
{
    int i;
    oident *ent;
    Z_DefaultDiagFormat *r;

    printf("Diagnostic message(s) from database:\n");
    for (i = 0; i<num; i++)
    {
        Z_DiagRec *p = pp[i];
        if (p->which != Z_DiagRec_defaultFormat)
        {
            printf("Diagnostic record not in default format.\n");
            return;
        }
        else
            r = p->u.defaultFormat;
        if (!(ent = oid_getentbyoid(r->diagnosticSetId)) ||
            ent->oclass != CLASS_DIAGSET || ent->value != VAL_BIB1)
            printf("Missing or unknown diagset\n");
        printf("    [%d] %s", *r->condition, diagbib1_str(*r->condition));
        switch (r->which)
        {
        case Z_DefaultDiagFormat_v2Addinfo:
            printf (" -- v2 addinfo '%s'\n", r->u.v2Addinfo);
            break;
        case Z_DefaultDiagFormat_v3Addinfo:
            printf (" -- v3 addinfo '%s'\n", r->u.v3Addinfo);
            break;
        }
    }
}


static void display_nameplusrecord(Z_NamePlusRecord *p)
{
    if (p->databaseName)
        printf("[%s]", p->databaseName);
    if (p->which == Z_NamePlusRecord_surrogateDiagnostic)
        display_diagrecs(&p->u.surrogateDiagnostic, 1);
    else if (p->which == Z_NamePlusRecord_databaseRecord)
        display_record(p->u.databaseRecord);
}

static void display_records(Z_Records *p)
{
    int i;

    if (p->which == Z_Records_NSD)
    {
        Z_DiagRec dr, *dr_p = &dr;
        dr.which = Z_DiagRec_defaultFormat;
        dr.u.defaultFormat = p->u.nonSurrogateDiagnostic;
        display_diagrecs (&dr_p, 1);
    }
    else if (p->which == Z_Records_multipleNSD)
        display_diagrecs (p->u.multipleNonSurDiagnostics->diagRecs,
                          p->u.multipleNonSurDiagnostics->num_diagRecs);
    else 
    {
        printf("Records: %d\n", p->u.databaseOrSurDiagnostics->num_records);
        for (i = 0; i < p->u.databaseOrSurDiagnostics->num_records; i++)
            display_nameplusrecord(p->u.databaseOrSurDiagnostics->records[i]);
    }
}

static int send_deleteResultSetRequest(char *arg)
{
    char names[8][32];
    int i;

    Z_APDU *apdu = zget_APDU(out, Z_APDU_deleteResultSetRequest);
    Z_DeleteResultSetRequest *req = apdu->u.deleteResultSetRequest;

    req->referenceId = set_refid (out);

    req->num_resultSetList =
        sscanf (arg, "%30s %30s %30s %30s %30s %30s %30s %30s",
                names[0], names[1], names[2], names[3],
                names[4], names[5], names[6], names[7]);

    req->deleteFunction = (int *)
        odr_malloc (out, sizeof(*req->deleteFunction));
    if (req->num_resultSetList > 0)
    {
        *req->deleteFunction = Z_DeleteRequest_list;
        req->resultSetList = (char **)
            odr_malloc (out, sizeof(*req->resultSetList)*
                        req->num_resultSetList);
        for (i = 0; i<req->num_resultSetList; i++)
            req->resultSetList[i] = names[i];
    }
    else
    {
        *req->deleteFunction = Z_DeleteRequest_all;
        req->resultSetList = 0;
    }
    
    send_apdu(apdu);
    printf("Sent deleteResultSetRequest.\n");
    return 2;
}

static int send_searchRequest(char *arg)
{
    Z_APDU *apdu = zget_APDU(out, Z_APDU_searchRequest);
    Z_SearchRequest *req = apdu->u.searchRequest;
    Z_Query query;
    int oid[OID_SIZE];
#if YAZ_MODULE_ccl
    struct ccl_rpn_node *rpn = NULL;
    int error, pos;
#endif
    char setstring[100];
    Z_RPNQuery *RPNquery;
    Odr_oct ccl_query;

#if YAZ_MODULE_ccl
    if (queryType == QueryType_CCL2RPN)
    {
        rpn = ccl_find_str(bibset, arg, &error, &pos);
        if (error)
        {
            printf("CCL ERROR: %s\n", ccl_err_msg(error));
            return 0;
        }
    }
#endif
    req->referenceId = set_refid (out);
    if (!strcmp(arg, "@big")) /* strictly for troublemaking */
    {
        static unsigned char big[2100];
        static Odr_oct bigo;

        /* send a very big referenceid to test transport stack etc. */
        memset(big, 'A', 2100);
        bigo.len = bigo.size = 2100;
        bigo.buf = big;
        req->referenceId = &bigo;
    }
    
    if (setnumber >= 0)
    {
        sprintf(setstring, "%d", ++setnumber);
        req->resultSetName = setstring;
    }
    *req->smallSetUpperBound = smallSetUpperBound;
    *req->largeSetLowerBound = largeSetLowerBound;
    *req->mediumSetPresentNumber = mediumSetPresentNumber;
    if (smallSetUpperBound > 0 || (largeSetLowerBound > 1 &&
        mediumSetPresentNumber > 0))
    {
        oident prefsyn;

        prefsyn.proto = protocol;
        prefsyn.oclass = CLASS_RECSYN;
        prefsyn.value = recordsyntax;
        req->preferredRecordSyntax =
            odr_oiddup(out, oid_ent_to_oid(&prefsyn, oid));
        req->smallSetElementSetNames =
            req->mediumSetElementSetNames = elementSetNames;
    }
    req->num_databaseNames = num_databaseNames;
    req->databaseNames = databaseNames;

    req->query = &query;

    switch (queryType)
    {
    case QueryType_Prefix:
        query.which = Z_Query_type_1;
        RPNquery = p_query_rpn (out, protocol, arg);
        if (!RPNquery)
        {
            printf("Prefix query error\n");
            return 0;
        }
        query.u.type_1 = RPNquery;
        break;
    case QueryType_CCL:
        query.which = Z_Query_type_2;
        query.u.type_2 = &ccl_query;
        ccl_query.buf = (unsigned char*) arg;
        ccl_query.len = strlen(arg);
        break;
#if YAZ_MODULE_ccl
    case QueryType_CCL2RPN:
        query.which = Z_Query_type_1;
        RPNquery = ccl_rpn_query(out, rpn);
        if (!RPNquery)
        {
            printf ("Couldn't convert from CCL to RPN\n");
            return 0;
        }
        query.u.type_1 = RPNquery;
        ccl_rpn_delete (rpn);
        break;
#endif
    default:
        printf ("Unsupported query type\n");
        return 0;
    }
    send_apdu(apdu);
    setno = 1;
    printf("Sent searchRequest.\n");
    return 2;
}

/* display Query Expression as part of searchResult-1 */
static void display_queryExpression (Z_QueryExpression *qe)
{
    if (!qe)
        return;
    if (qe->which == Z_QueryExpression_term)
    {
        if (qe->u.term->queryTerm)
        {
            Z_Term *term = qe->u.term->queryTerm;
            if (term->which == Z_Term_general)
                printf (" %.*s", term->u.general->len, term->u.general->buf);
        }
    }

}

/* see if we can find USR:SearchResult-1 */
static void display_searchResult (Z_OtherInformation *o)
{
    int i;
    if (!o)
        return ;
    for (i = 0; i < o->num_elements; i++)
    {
        if (o->list[i]->which == Z_OtherInfo_externallyDefinedInfo)
        {
            Z_External *ext = o->list[i]->information.externallyDefinedInfo;
            
            if (ext->which == Z_External_searchResult1)
            {
                int j;
                Z_SearchInfoReport *sr = ext->u.searchResult1;
                printf ("SearchResult-1:");
                for (j = 0; j < sr->num; j++)
                {
                    if (!sr->elements[j]->subqueryExpression)
                        printf (" %d", j);
                    display_queryExpression (
                        sr->elements[j]->subqueryExpression);
                    display_queryExpression (
                        sr->elements[j]->subqueryInterpretation);
                    display_queryExpression (
                        sr->elements[j]->subqueryRecommendation);
                    if (sr->elements[j]->subqueryCount)
                        printf ("(%d)", *sr->elements[j]->subqueryCount);
                }
                printf ("\n");
            }
        }
    }
}

static int process_searchResponse(Z_SearchResponse *res)
{
    printf ("Received SearchResponse.\n");
    print_refid (res->referenceId);
    if (*res->searchStatus)
        printf("Search was a success.\n");
    else
        printf("Search was a bloomin' failure.\n");
    printf("Number of hits: %d", *res->resultCount);
    if (setnumber >= 0)
        printf (", setno %d", setnumber);
    printf ("\n");
    display_searchResult (res->additionalSearchInfo);
    printf("records returned: %d\n",
           *res->numberOfRecordsReturned);
    setno += *res->numberOfRecordsReturned;
    if (res->records)
        display_records(res->records);
    return 0;
}

static void print_level(int iLevel)
{
    int i;
    for (i = 0; i < iLevel * 4; i++)
        printf(" ");
}

static void print_int(int iLevel, const char *pTag, int *pInt)
{
    if (pInt != NULL)
    {
        print_level(iLevel);
        printf("%s: %d\n", pTag, *pInt);
    }
}

static void print_string(int iLevel, const char *pTag, const char *pString)
{
    if (pString != NULL)
    {
        print_level(iLevel);
        printf("%s: %s\n", pTag, pString);
    }
}

static void print_oid(int iLevel, const char *pTag, Odr_oid *pOid)
{
    if (pOid != NULL)
    {
        int *pInt = pOid;

        print_level(iLevel);
        printf("%s:", pTag);
        for (; *pInt != -1; pInt++)
            printf(" %d", *pInt);
        printf("\n");
    }
}

static void print_referenceId(int iLevel, Z_ReferenceId *referenceId)
{
    if (referenceId != NULL)
    {
        int i;

        print_level(iLevel);
        printf("Ref Id (%d, %d): ", referenceId->len, referenceId->size);
        for (i = 0; i < referenceId->len; i++)
            printf("%c", referenceId->buf[i]);
        printf("\n");
    }
}

static void print_string_or_numeric(int iLevel, const char *pTag, Z_StringOrNumeric *pStringNumeric)
{
    if (pStringNumeric != NULL)
    {
        switch (pStringNumeric->which)
        {
        case Z_StringOrNumeric_string:
            print_string(iLevel, pTag, pStringNumeric->u.string);
            break;
            
        case Z_StringOrNumeric_numeric:
            print_int(iLevel, pTag, pStringNumeric->u.numeric);
            break;
            
        default:
            print_level(iLevel);
            printf("%s: valid type for Z_StringOrNumeric\n", pTag);
            break;
        }
    }
}

static void print_universe_report_duplicate(int iLevel, Z_UniverseReportDuplicate *pUniverseReportDuplicate)
{
    if (pUniverseReportDuplicate != NULL)
    {
        print_level(iLevel);
        printf("Universe Report Duplicate: \n");
        iLevel++;
        print_string_or_numeric(iLevel, "Hit No",
                                pUniverseReportDuplicate->hitno);
    }
}

static void
print_universe_report_hits(int iLevel,
                           Z_UniverseReportHits *pUniverseReportHits)
{
    if (pUniverseReportHits != NULL)
    {
        print_level(iLevel);
        printf("Universe Report Hits: \n");
        iLevel++;
        print_string_or_numeric(iLevel, "Database",
                                pUniverseReportHits->database);
        print_string_or_numeric(iLevel, "Hits", pUniverseReportHits->hits);
    }
}

static void print_universe_report(int iLevel, Z_UniverseReport *pUniverseReport)
{
    if (pUniverseReport != NULL)
    {
        print_level(iLevel);
        printf("Universe Report: \n");
        iLevel++;
        print_int(iLevel, "Total Hits", pUniverseReport->totalHits);
        switch (pUniverseReport->which)
        {
        case Z_UniverseReport_databaseHits:
            print_universe_report_hits(iLevel, pUniverseReport->u.databaseHits);
            break;
            
        case Z_UniverseReport_duplicate:
            print_universe_report_duplicate(iLevel, pUniverseReport->u.duplicate);
            break;
            
        default:
            print_level(iLevel);
            printf("Type: %d\n", pUniverseReport->which);
            break;
        }
    }
}

static void print_external(int iLevel, Z_External *pExternal)
{
    if (pExternal != NULL)
    {
        print_level(iLevel);
        printf("External: \n");
        iLevel++;
        print_oid(iLevel, "Direct Reference", pExternal->direct_reference);
        print_int(iLevel, "InDirect Reference", pExternal->indirect_reference);
        print_string(iLevel, "Descriptor", pExternal->descriptor);
        switch (pExternal->which)
        {
        case Z_External_universeReport:
            print_universe_report(iLevel, pExternal->u.universeReport);
            break;
            
        default:
            print_level(iLevel);
            printf("Type: %d\n", pExternal->which);
            break;
        }
    }
}

static int process_resourceControlRequest (Z_ResourceControlRequest *req)
{
    printf ("Received ResourceControlRequest.\n");
    print_referenceId(1, req->referenceId);
    print_int(1, "Suspended Flag", req->suspendedFlag);
    print_int(1, "Partial Results Available", req->partialResultsAvailable);
    print_int(1, "Response Required", req->responseRequired);
    print_int(1, "Triggered Request Flag", req->triggeredRequestFlag);
    print_external(1, req->resourceReport);
    return 0;
}

void process_ESResponse(Z_ExtendedServicesResponse *res)
{
    printf("Status: ");
    switch (*res->operationStatus)
    {
    case Z_ExtendedServicesResponse_done:
        printf ("done\n");
        break;
    case Z_ExtendedServicesResponse_accepted:
        printf ("accepted\n");
        break;
    case Z_ExtendedServicesResponse_failure:
        printf ("failure\n");
        display_diagrecs(res->diagnostics, res->num_diagnostics);
        break;
    default:
        printf ("unknown\n");
    }
    if ( (*res->operationStatus != Z_ExtendedServicesResponse_failure) &&
        (res->num_diagnostics != 0) ) {
        display_diagrecs(res->diagnostics, res->num_diagnostics);
    }
    print_refid (res->referenceId);
    if (res->taskPackage && 
        res->taskPackage->which == Z_External_extendedService)
    {
        Z_TaskPackage *taskPackage = res->taskPackage->u.extendedService;
        Odr_oct *id = taskPackage->targetReference;
        Z_External *ext = taskPackage->taskSpecificParameters;
        
        if (id)
        {
            printf ("Target Reference: ");
            print_stringn (id->buf, id->len);
            printf ("\n");
        }
        if (ext->which == Z_External_update)
        {
            Z_IUUpdateTaskPackage *utp = ext->u.update->u.taskPackage;
            if (utp && utp->targetPart)
            {
                Z_IUTargetPart *targetPart = utp->targetPart;
                int i;

                for (i = 0; i<targetPart->num_taskPackageRecords;  i++)
                {

                    Z_IUTaskPackageRecordStructure *tpr =
                        targetPart->taskPackageRecords[i];
                    printf ("task package record %d\n", i+1);
                    if (tpr->which == Z_IUTaskPackageRecordStructure_record)
                    {
                        display_record (tpr->u.record);
                    }
                    else
                    {
                        printf ("other type\n");
                    }
                }
            }
        }
    }
}

#if YAZ_MODULE_ill

const char *get_ill_element (void *clientData, const char *element)
{
    return 0;
}

static Z_External *create_external_itemRequest()
{
    struct ill_get_ctl ctl;
    ILL_ItemRequest *req;
    Z_External *r = 0;
    int item_request_size = 0;
    char *item_request_buf = 0;

    ctl.odr = out;
    ctl.clientData = 0;
    ctl.f = get_ill_element;
    
    req = ill_get_ItemRequest(&ctl, "ill", 0);
    if (!req)
        printf ("ill_get_ItemRequest failed\n");
        
    if (!ill_ItemRequest (out, &req, 0, 0))
    {
        if (apdu_file)
        {
            ill_ItemRequest(print, &req, 0, 0);
            odr_reset(print);
        }
        item_request_buf = odr_getbuf (out, &item_request_size, 0);
        if (item_request_buf)
            odr_setbuf (out, item_request_buf, item_request_size, 1);
        printf ("Couldn't encode ItemRequest, size %d\n", item_request_size);
        return 0;
    }
    else
    {
        oident oid;
        
        item_request_buf = odr_getbuf (out, &item_request_size, 0);
        oid.proto = PROTO_GENERAL;
        oid.oclass = CLASS_GENERAL;
        oid.value = VAL_ISO_ILL_1;
        
        r = (Z_External *) odr_malloc (out, sizeof(*r));
        r->direct_reference = odr_oiddup(out,oid_getoidbyent(&oid)); 
        r->indirect_reference = 0;
        r->descriptor = 0;
        r->which = Z_External_single;
        
        r->u.single_ASN1_type = (Odr_oct *)
            odr_malloc (out, sizeof(*r->u.single_ASN1_type));
        r->u.single_ASN1_type->buf = (unsigned char *)
        odr_malloc (out, item_request_size);
        r->u.single_ASN1_type->len = item_request_size;
        r->u.single_ASN1_type->size = item_request_size;
        memcpy (r->u.single_ASN1_type->buf, item_request_buf,
                item_request_size);
        printf ("len = %d\n", item_request_size);
    }
    return r;
}
#endif

#ifdef YAZ_MODULE_ill
static Z_External *create_external_ILL_APDU(int which)
{
    struct ill_get_ctl ctl;
    ILL_APDU *ill_apdu;
    Z_External *r = 0;
    int ill_request_size = 0;
    char *ill_request_buf = 0;
        
    ctl.odr = out;
    ctl.clientData = 0;
    ctl.f = get_ill_element;

    ill_apdu = ill_get_APDU(&ctl, "ill", 0);

    if (!ill_APDU (out, &ill_apdu, 0, 0))
    {
        if (apdu_file)
        {
            printf ("-------------------\n");
            ill_APDU(print, &ill_apdu, 0, 0);
            odr_reset(print);
            printf ("-------------------\n");
        }
        ill_request_buf = odr_getbuf (out, &ill_request_size, 0);
        if (ill_request_buf)
            odr_setbuf (out, ill_request_buf, ill_request_size, 1);
        printf ("Couldn't encode ILL-Request, size %d\n", ill_request_size);
        return 0;
    }
    else
    {
        oident oid;
        ill_request_buf = odr_getbuf (out, &ill_request_size, 0);
        
        oid.proto = PROTO_GENERAL;
        oid.oclass = CLASS_GENERAL;
        oid.value = VAL_ISO_ILL_1;
        
        r = (Z_External *) odr_malloc (out, sizeof(*r));
        r->direct_reference = odr_oiddup(out,oid_getoidbyent(&oid)); 
        r->indirect_reference = 0;
        r->descriptor = 0;
        r->which = Z_External_single;
        
        r->u.single_ASN1_type = (Odr_oct *)
            odr_malloc (out, sizeof(*r->u.single_ASN1_type));
        r->u.single_ASN1_type->buf = (unsigned char *)
        odr_malloc (out, ill_request_size);
        r->u.single_ASN1_type->len = ill_request_size;
        r->u.single_ASN1_type->size = ill_request_size;
        memcpy (r->u.single_ASN1_type->buf, ill_request_buf, ill_request_size);
        printf ("len = %d\n", ill_request_size);
    }
    return r;
}
#endif


static Z_External *create_ItemOrderExternal(const char *type, int itemno)
{
    Z_External *r = (Z_External *) odr_malloc(out, sizeof(Z_External));
    oident ItemOrderRequest;
  
    ItemOrderRequest.proto = PROTO_Z3950;
    ItemOrderRequest.oclass = CLASS_EXTSERV;
    ItemOrderRequest.value = VAL_ITEMORDER;
 
    r->direct_reference = odr_oiddup(out,oid_getoidbyent(&ItemOrderRequest)); 
    r->indirect_reference = 0;
    r->descriptor = 0;

    r->which = Z_External_itemOrder;

    r->u.itemOrder = (Z_ItemOrder *) odr_malloc(out,sizeof(Z_ItemOrder));
    memset(r->u.itemOrder, 0, sizeof(Z_ItemOrder));
    r->u.itemOrder->which=Z_IOItemOrder_esRequest;

    r->u.itemOrder->u.esRequest = (Z_IORequest *) 
        odr_malloc(out,sizeof(Z_IORequest));
    memset(r->u.itemOrder->u.esRequest, 0, sizeof(Z_IORequest));

    r->u.itemOrder->u.esRequest->toKeep = (Z_IOOriginPartToKeep *)
        odr_malloc(out,sizeof(Z_IOOriginPartToKeep));
    memset(r->u.itemOrder->u.esRequest->toKeep, 0, sizeof(Z_IOOriginPartToKeep));
    r->u.itemOrder->u.esRequest->notToKeep = (Z_IOOriginPartNotToKeep *)
        odr_malloc(out,sizeof(Z_IOOriginPartNotToKeep));
    memset(r->u.itemOrder->u.esRequest->notToKeep, 0, sizeof(Z_IOOriginPartNotToKeep));

    r->u.itemOrder->u.esRequest->toKeep->supplDescription = NULL;
    r->u.itemOrder->u.esRequest->toKeep->contact = NULL;
    r->u.itemOrder->u.esRequest->toKeep->addlBilling = NULL;

    r->u.itemOrder->u.esRequest->notToKeep->resultSetItem =
        (Z_IOResultSetItem *) odr_malloc(out, sizeof(Z_IOResultSetItem));
    memset(r->u.itemOrder->u.esRequest->notToKeep->resultSetItem, 0, sizeof(Z_IOResultSetItem));
    r->u.itemOrder->u.esRequest->notToKeep->resultSetItem->resultSetId = "1";

    r->u.itemOrder->u.esRequest->notToKeep->resultSetItem->item =
        (int *) odr_malloc(out, sizeof(int));
    *r->u.itemOrder->u.esRequest->notToKeep->resultSetItem->item = itemno;

#if YAZ_MODULE_ill
    if (!strcmp (type, "item") || !strcmp(type, "2"))
    {
        printf ("using item-request\n");
        r->u.itemOrder->u.esRequest->notToKeep->itemRequest = 
            create_external_itemRequest();
    }
    else if (!strcmp(type, "ill") || !strcmp(type, "1"))
    {
        printf ("using ILL-request\n");
        r->u.itemOrder->u.esRequest->notToKeep->itemRequest = 
            create_external_ILL_APDU(ILL_APDU_ILL_Request);
    }
    else if (!strcmp(type, "xml") || !strcmp(type, "3"))
    {
    const char *xml_buf =
        "<itemorder>\n"
        "  <type>request</type>\n"
        "  <libraryNo>000200</libraryNo>\n"
        "  <borrowerTicketNo> 1212 </borrowerTicketNo>\n"
        "</itemorder>";
        r->u.itemOrder->u.esRequest->notToKeep->itemRequest =
            z_ext_record (out, VAL_TEXT_XML, xml_buf, strlen(xml_buf));
    }
    else
        r->u.itemOrder->u.esRequest->notToKeep->itemRequest = 0;

#else
    r->u.itemOrder->u.esRequest->notToKeep->itemRequest = 0;
#endif
    return r;
}

static int send_itemorder(const char *type, int itemno)
{
    Z_APDU *apdu = zget_APDU(out, Z_APDU_extendedServicesRequest);
    Z_ExtendedServicesRequest *req = apdu->u.extendedServicesRequest;
    oident ItemOrderRequest;

    ItemOrderRequest.proto = PROTO_Z3950;
    ItemOrderRequest.oclass = CLASS_EXTSERV;
    ItemOrderRequest.value = VAL_ITEMORDER;
    req->packageType = odr_oiddup(out,oid_getoidbyent(&ItemOrderRequest));
    req->packageName = esPackageName;

    req->taskSpecificParameters = create_ItemOrderExternal(type, itemno);

    send_apdu(apdu);
    return 0;
}

static int cmd_update(char *arg)
{
    Z_APDU *apdu = zget_APDU(out, Z_APDU_extendedServicesRequest );
    Z_ExtendedServicesRequest *req = apdu->u.extendedServicesRequest;
    Z_External *r;
    int oid[OID_SIZE];
    Z_IUOriginPartToKeep *toKeep;
    Z_IUSuppliedRecords *notToKeep;
    oident update_oid;
    printf ("Update request\n");
    fflush(stdout);
    
    if (!record_last)
        return 0;
    update_oid.proto = PROTO_Z3950;
    update_oid.oclass = CLASS_EXTSERV;
    update_oid.value = VAL_DBUPDATE;
    oid_ent_to_oid (&update_oid, oid);
    req->packageType = odr_oiddup(out,oid);
    req->packageName = esPackageName;
    
    req->referenceId = set_refid (out);

    r = req->taskSpecificParameters = (Z_External *)
        odr_malloc (out, sizeof(*r));
    r->direct_reference = odr_oiddup(out,oid);
    r->indirect_reference = 0;
    r->descriptor = 0;
    r->which = Z_External_update;
    r->u.update = (Z_IUUpdate *) odr_malloc(out, sizeof(*r->u.update));
    r->u.update->which = Z_IUUpdate_esRequest;
    r->u.update->u.esRequest = (Z_IUUpdateEsRequest *)
        odr_malloc(out, sizeof(*r->u.update->u.esRequest));
    toKeep = r->u.update->u.esRequest->toKeep = (Z_IUOriginPartToKeep *)
        odr_malloc(out, sizeof(*r->u.update->u.esRequest->toKeep));
    toKeep->databaseName = databaseNames[0];
    toKeep->schema = 0;
    toKeep->elementSetName = 0;
    toKeep->actionQualifier = 0;
    toKeep->action = (int *) odr_malloc(out, sizeof(*toKeep->action));
    *toKeep->action = Z_IUOriginPartToKeep_recordInsert;

    notToKeep = r->u.update->u.esRequest->notToKeep = (Z_IUSuppliedRecords *)
        odr_malloc(out, sizeof(*r->u.update->u.esRequest->notToKeep));
    notToKeep->num = 1;
    notToKeep->elements = (Z_IUSuppliedRecords_elem **)
        odr_malloc(out, sizeof(*notToKeep->elements));
    notToKeep->elements[0] = (Z_IUSuppliedRecords_elem *)
        odr_malloc(out, sizeof(**notToKeep->elements));
    notToKeep->elements[0]->u.number = 0;
    notToKeep->elements[0]->supplementalId = 0;
    notToKeep->elements[0]->correlationInfo = 0;
    notToKeep->elements[0]->record = record_last;
    
    send_apdu(apdu);

    return 2;
}

static int cmd_itemorder(char *arg)
{
    char type[12];
    int itemno;
    
    if (sscanf (arg, "%10s %d", type, &itemno) != 2)
        return 0;

    printf("Item order request\n");
    fflush(stdout);
    send_itemorder(type, itemno);
    return 2;
}

static int cmd_find(char *arg)
{
    if (!*arg)
    {
        printf("Find what?\n");
        return 0;
    }
    if (!conn)
    {
        printf("Not connected yet\n");
        return 0;
    }
    if (!send_searchRequest(arg))
        return 0;
    return 2;
}

static int cmd_delete(char *arg)
{
    if (!conn)
    {
        printf("Not connected yet\n");
        return 0;
    }
    if (!send_deleteResultSetRequest(arg))
        return 0;
    return 2;
}

static int cmd_ssub(char *arg)
{
    if (!(smallSetUpperBound = atoi(arg)))
        return 0;
    return 1;
}

static int cmd_lslb(char *arg)
{
    if (!(largeSetLowerBound = atoi(arg)))
        return 0;
    return 1;
}

static int cmd_mspn(char *arg)
{
    if (!(mediumSetPresentNumber = atoi(arg)))
        return 0;
    return 1;
}

static int cmd_status(char *arg)
{
    printf("smallSetUpperBound: %d\n", smallSetUpperBound);
    printf("largeSetLowerBound: %d\n", largeSetLowerBound);
    printf("mediumSetPresentNumber: %d\n", mediumSetPresentNumber);
    return 1;
}

static int cmd_setnames(char *arg)
{
    if (*arg == '1')         /* enable ? */
        setnumber = 0;
    else if (*arg == '0')    /* disable ? */
        setnumber = -1;
    else if (setnumber < 0)  /* no args, toggle .. */
        setnumber = 0;
    else
        setnumber = -1;
   
    if (setnumber >= 0)
        printf("Set numbering enabled.\n");
    else
        printf("Set numbering disabled.\n");
    return 1;
}

/* PRESENT SERVICE ----------------------------- */

static int send_presentRequest(char *arg)
{
    Z_APDU *apdu = zget_APDU(out, Z_APDU_presentRequest);
    Z_PresentRequest *req = apdu->u.presentRequest;
    Z_RecordComposition compo;
    oident prefsyn;
    int nos = 1;
    int oid[OID_SIZE];
    char *p;
    char setstring[100];

    req->referenceId = set_refid (out);
    if ((p = strchr(arg, '+')))
    {
        nos = atoi(p + 1);
        *p = 0;
    }
    if (*arg)
        setno = atoi(arg);
    if (p && (p=strchr(p+1, '+')))
    {
        strcpy (setstring, p+1);
        req->resultSetId = setstring;
    }
    else if (setnumber >= 0)
    {
        sprintf(setstring, "%d", setnumber);
        req->resultSetId = setstring;
    }
    req->resultSetStartPoint = &setno;
    req->numberOfRecordsRequested = &nos;
    prefsyn.proto = protocol;
    prefsyn.oclass = CLASS_RECSYN;
    prefsyn.value = recordsyntax;
    req->preferredRecordSyntax =
        odr_oiddup (out, oid_ent_to_oid(&prefsyn, oid));

    if (schema != VAL_NONE)
    {
        oident prefschema;

        prefschema.proto = protocol;
        prefschema.oclass = CLASS_SCHEMA;
        prefschema.value = schema;

        req->recordComposition = &compo;
        compo.which = Z_RecordComp_complex;
        compo.u.complex = (Z_CompSpec *)
            odr_malloc(out, sizeof(*compo.u.complex));
        compo.u.complex->selectAlternativeSyntax = (bool_t *) 
            odr_malloc(out, sizeof(bool_t));
        *compo.u.complex->selectAlternativeSyntax = 0;

        compo.u.complex->generic = (Z_Specification *)
            odr_malloc(out, sizeof(*compo.u.complex->generic));
        compo.u.complex->generic->schema = (Odr_oid *)
            odr_oiddup(out, oid_ent_to_oid(&prefschema, oid));
        if (!compo.u.complex->generic->schema)
        {
            /* OID wasn't a schema! Try record syntax instead. */
            prefschema.oclass = CLASS_RECSYN;
            compo.u.complex->generic->schema = (Odr_oid *)
                odr_oiddup(out, oid_ent_to_oid(&prefschema, oid));
        }
        if (!elementSetNames)
            compo.u.complex->generic->elementSpec = 0;
        else
        {
            compo.u.complex->generic->elementSpec = (Z_ElementSpec *)
                odr_malloc(out, sizeof(Z_ElementSpec));
            compo.u.complex->generic->elementSpec->which =
                Z_ElementSpec_elementSetName;
            compo.u.complex->generic->elementSpec->u.elementSetName =
                elementSetNames->u.generic;
        }
        compo.u.complex->num_dbSpecific = 0;
        compo.u.complex->dbSpecific = 0;
        compo.u.complex->num_recordSyntax = 0;
        compo.u.complex->recordSyntax = 0;
    }
    else if (elementSetNames)
    {
        req->recordComposition = &compo;
        compo.which = Z_RecordComp_simple;
        compo.u.simple = elementSetNames;
    }
    send_apdu(apdu);
    printf("Sent presentRequest (%d+%d).\n", setno, nos);
    return 2;
}
    
static void close_session (void)
{
    cs_close (conn);
    conn = 0;
    if (session_mem)
    {
        nmem_destroy (session_mem);
        session_mem = NULL;
    }
    sent_close = 0;
}

void process_close(Z_Close *req)
{
    Z_APDU *apdu = zget_APDU(out, Z_APDU_close);
    Z_Close *res = apdu->u.close;

    static char *reasons[] =
    {
        "finished",
        "shutdown",
        "system problem",
        "cost limit reached",
        "resources",
        "security violation",
        "protocolError",
        "lack of activity",
        "peer abort",
        "unspecified"
    };

    printf("Reason: %s, message: %s\n", reasons[*req->closeReason],
        req->diagnosticInformation ? req->diagnosticInformation : "NULL");
    if (sent_close)
        close_session ();
    else
    {
        *res->closeReason = Z_Close_finished;
        send_apdu(apdu);
        printf("Sent response.\n");
        sent_close = 1;
    }
}

static int cmd_show(char *arg)
{
    if (!conn)
    {
        printf("Not connected yet\n");
        return 0;
    }
    if (!send_presentRequest(arg))
        return 0;
    return 2;
}

int cmd_quit(char *arg)
{
    printf("See you later, alligator.\n");
    xmalloc_trav ("");
    exit(0);
    return 0;
}

int cmd_cancel(char *arg)
{
    Z_APDU *apdu = zget_APDU(out, Z_APDU_triggerResourceControlRequest);
    Z_TriggerResourceControlRequest *req =
        apdu->u.triggerResourceControlRequest;
    bool_t rfalse = 0;
    
    if (!conn)
    {
        printf("Session not initialized yet\n");
        return 0;
    }
    if (!ODR_MASK_GET(session->options, Z_Options_triggerResourceCtrl))
    {
        printf("Target doesn't support cancel (trigger resource ctrl)\n");
        return 0;
    }
    *req->requestedAction = Z_TriggerResourceCtrl_cancel;
    req->resultSetWanted = &rfalse;

    send_apdu(apdu);
    printf("Sent cancel request\n");
    return 2;
}

int send_scanrequest(const char *query, int pp, int num, const char *term)
{
    Z_APDU *apdu = zget_APDU(out, Z_APDU_scanRequest);
    Z_ScanRequest *req = apdu->u.scanRequest;
    int use_rpn = 1;
#if YAZ_MODULE_ccl
    int oid[OID_SIZE];
    
    if (queryType == QueryType_CCL2RPN)
    {
        oident bib1;
        int error, pos;
        struct ccl_rpn_node *rpn;

        rpn = ccl_find_str (bibset,  query, &error, &pos);
        if (error)
        {
            printf("CCL ERROR: %s\n", ccl_err_msg(error));
            return -1;
        }
        use_rpn = 0;
        bib1.proto = PROTO_Z3950;
        bib1.oclass = CLASS_ATTSET;
        bib1.value = VAL_BIB1;
        req->attributeSet = oid_ent_to_oid (&bib1, oid);
        if (!(req->termListAndStartPoint = ccl_scan_query (out, rpn)))
        {
            printf("Couldn't convert CCL to Scan term\n");
            return -1;
        }
        ccl_rpn_delete (rpn);
    }
#endif
    if (use_rpn && !(req->termListAndStartPoint =
                     p_query_scan(out, protocol, &req->attributeSet, query)))
    {
        printf("Prefix query error\n");
        return -1;
    }
    if (term && *term)
    {
        if (req->termListAndStartPoint->term &&
            req->termListAndStartPoint->term->which == Z_Term_general &&
            req->termListAndStartPoint->term->u.general)
        {
            req->termListAndStartPoint->term->u.general->buf =
                (unsigned char *) odr_strdup(out, term);
            req->termListAndStartPoint->term->u.general->len =
                req->termListAndStartPoint->term->u.general->size =
                strlen(term);
        }
    }
    req->referenceId = set_refid (out);
    req->num_databaseNames = num_databaseNames;
    req->databaseNames = databaseNames;
    req->numberOfTermsRequested = &num;
    req->preferredPositionInResponse = &pp;
    send_apdu(apdu);
    return 2;
}

int send_sortrequest(char *arg, int newset)
{
    Z_APDU *apdu = zget_APDU(out, Z_APDU_sortRequest);
    Z_SortRequest *req = apdu->u.sortRequest;
    Z_SortKeySpecList *sksl = (Z_SortKeySpecList *)
        odr_malloc (out, sizeof(*sksl));
    char setstring[32];

    if (setnumber >= 0)
        sprintf (setstring, "%d", setnumber);
    else
        sprintf (setstring, "default");

    req->referenceId = set_refid (out);

    req->num_inputResultSetNames = 1;
    req->inputResultSetNames = (Z_InternationalString **)
        odr_malloc (out, sizeof(*req->inputResultSetNames));
    req->inputResultSetNames[0] = odr_strdup (out, setstring);

    if (newset && setnumber >= 0)
        sprintf (setstring, "%d", ++setnumber);

    req->sortedResultSetName = odr_strdup (out, setstring);

    req->sortSequence = yaz_sort_spec (out, arg);
    if (!req->sortSequence)
    {
        printf ("Missing sort specifications\n");
        return -1;
    }
    send_apdu(apdu);
    return 2;
}

void display_term(Z_TermInfo *t)
{
    if (t->term->which == Z_Term_general)
    {
        printf("%.*s", t->term->u.general->len, t->term->u.general->buf);
        sprintf(last_scan_line, "%.*s", t->term->u.general->len,
            t->term->u.general->buf);
    }
    else
        printf("Term (not general)");
    if (t->globalOccurrences)
        printf (" (%d)\n", *t->globalOccurrences);
    else
        printf ("\n");
}

void process_scanResponse(Z_ScanResponse *res)
{
    int i;
    Z_Entry **entries = NULL;
    int num_entries = 0;
   
    printf("Received ScanResponse\n"); 
    print_refid (res->referenceId);
    printf("%d entries", *res->numberOfEntriesReturned);
    if (res->positionOfTerm)
        printf (", position=%d", *res->positionOfTerm); 
    printf ("\n");
    if (*res->scanStatus != Z_Scan_success)
        printf("Scan returned code %d\n", *res->scanStatus);
    if (!res->entries)
        return;
    if ((entries = res->entries->entries))
        num_entries = res->entries->num_entries;
    for (i = 0; i < num_entries; i++)
    {
        int pos_term = res->positionOfTerm ? *res->positionOfTerm : -1;
        if (entries[i]->which == Z_Entry_termInfo)
        {
            printf("%c ", i + 1 == pos_term ? '*' : ' ');
            display_term(entries[i]->u.termInfo);
        }
        else
            display_diagrecs(&entries[i]->u.surrogateDiagnostic, 1);
    }
    if (res->entries->nonsurrogateDiagnostics)
        display_diagrecs (res->entries->nonsurrogateDiagnostics,
                          res->entries->num_nonsurrogateDiagnostics);
}

void process_sortResponse(Z_SortResponse *res)
{
    printf("Received SortResponse: status=");
    switch (*res->sortStatus)
    {
    case Z_SortStatus_success:
        printf ("success"); break;
    case Z_SortStatus_partial_1:
        printf ("partial"); break;
    case Z_SortStatus_failure:
        printf ("failure"); break;
    default:
        printf ("unknown (%d)", *res->sortStatus);
    }
    printf ("\n");
    print_refid (res->referenceId);
    if (res->diagnostics)
        display_diagrecs(res->diagnostics,
                         res->num_diagnostics);
}

void process_deleteResultSetResponse (Z_DeleteResultSetResponse *res)
{
    printf("Got deleteResultSetResponse status=%d\n",
           *res->deleteOperationStatus);
    if (res->deleteListStatuses)
    {
        int i;
        for (i = 0; i < res->deleteListStatuses->num; i++)
        {
            printf ("%s status=%d\n", res->deleteListStatuses->elements[i]->id,
                    *res->deleteListStatuses->elements[i]->status);
        }
    }
}

int cmd_sort_generic(char *arg, int newset)
{
    if (!conn)
    {
        printf("Session not initialized yet\n");
        return 0;
    }
    if (!ODR_MASK_GET(session->options, Z_Options_sort))
    {
        printf("Target doesn't support sort\n");
        return 0;
    }
    if (*arg)
    {
        if (send_sortrequest(arg, newset) < 0)
            return 0;
        return 2;
    }
    return 0;
}

int cmd_sort(char *arg)
{
    return cmd_sort_generic (arg, 0);
}

int cmd_sort_newset (char *arg)
{
    return cmd_sort_generic (arg, 1);
}

int cmd_scan(char *arg)
{
    if (!conn)
    {
        printf("Session not initialized yet\n");
        return 0;
    }
    if (!ODR_MASK_GET(session->options, Z_Options_scan))
    {
        printf("Target doesn't support scan\n");
        return 0;
    }
    if (*arg)
    {
        strcpy (last_scan_query, arg);
        if (send_scanrequest(arg, 1, 20, 0) < 0)
            return 0;
    }
    else
    {
        if (send_scanrequest(last_scan_query, 1, 20, last_scan_line) < 0)
            return 0;
    }
    return 2;
}

int cmd_schema(char *arg)
{
    if (!arg || !*arg)
    {
        schema = VAL_NONE;
        return 1;
    }
    schema = oid_getvalbyname (arg);
    if (schema == VAL_NONE)
    {
        printf ("unknown schema\n");
        return 0;
    }
    return 1;
}

int cmd_format(char *arg)
{
    if (!arg || !*arg)
    {
        printf("Usage: format <recordsyntax>\n");
        return 0;
    }
    recordsyntax = oid_getvalbyname (arg);
    if (recordsyntax == VAL_NONE)
    {
        printf ("unknown record syntax\n");
        return 0;
    }
    return 1;
}

int cmd_elements(char *arg)
{
    static Z_ElementSetNames esn;
    static char what[100];

    if (!arg || !*arg)
    {
        elementSetNames = 0;
        return 1;
    }
    strcpy(what, arg);
    esn.which = Z_ElementSetNames_generic;
    esn.u.generic = what;
    elementSetNames = &esn;
    return 1;
}

int cmd_attributeset(char *arg)
{
    char what[100];

    if (!arg || !*arg)
    {
        printf("Usage: attributeset <setname>\n");
        return 0;
    }
    sscanf(arg, "%s", what);
    if (p_query_attset (what))
    {
        printf("Unknown attribute set name\n");
        return 0;
    }
    return 1;
}

int cmd_querytype (char *arg)
{
    if (!strcmp (arg, "ccl"))
        queryType = QueryType_CCL;
    else if (!strcmp (arg, "prefix") || !strcmp(arg, "rpn"))
        queryType = QueryType_Prefix;
#if YAZ_MODULE_ccl
    else if (!strcmp (arg, "ccl2rpn") || !strcmp (arg, "cclrpn"))
        queryType = QueryType_CCL2RPN;
#endif
    else
    {
        printf ("Querytype must be one of:\n");
        printf (" prefix         - Prefix query\n");
        printf (" ccl            - CCL query\n");
#if YAZ_MODULE_ccl
        printf (" ccl2rpn        - CCL query converted to RPN\n");
#endif
        return 0;
    }
    return 1;
}

int cmd_refid (char *arg)
{
    xfree (refid);
    refid = NULL;
    if (*arg)
    {
        refid = (char *) xmalloc (strlen(arg)+1);
        strcpy (refid, arg);
    }
    return 1;
}

int cmd_close(char *arg)
{
    Z_APDU *apdu;
    Z_Close *req;
    if (!conn)
        return 0;

    apdu = zget_APDU(out, Z_APDU_close);
    req = apdu->u.close;
    *req->closeReason = Z_Close_finished;
    send_apdu(apdu);
    printf("Sent close request.\n");
    sent_close = 1;
    return 2;
}

int cmd_packagename(char* arg)
{
    xfree (esPackageName);
    esPackageName = NULL;
    if (*arg)
    {
        esPackageName = (char *) xmalloc (strlen(arg)+1);
        strcpy (esPackageName, arg);
    }
    return 1;
}

int cmd_proxy(char* arg)
{
    xfree (yazProxy);
    yazProxy = NULL;
    if (*arg)
    {
        yazProxy = (char *) xmalloc (strlen(arg)+1);
        strcpy (yazProxy, arg);
    } 
    return 1;
}

int cmd_source(char* arg) 
{
    /* first should open the file and read one line at a time.. */
    FILE* includeFile;
    char line[1024], *cp;

    {
        char* args_end=(arg)+strlen(arg)-1; 
        while(isspace(*args_end)) 
        {*args_end=0;
        --args_end;}; 
    }

    REMOVE_TAILING_BLANKS(arg);
    
    if(strlen(arg)<1) {
        fprintf(stderr,"Error in source command use a filename\n");
        return -1;
    }
    
    includeFile = fopen (arg, "r");
    
    if(!includeFile) {
        fprintf(stderr,"Unable to open file %s for reading\n",arg);
        return -1;
    }
    
    while(!feof(includeFile)) {
        memset(line,0,sizeof(line));
        fgets(line,sizeof(line),includeFile);
        
        if(strlen(line) < 2) continue;
        if(line[0] == '#') continue;
        
        if ((cp = strrchr (line, '\n')))
            *cp = '\0';
        
        process_cmd_line(line);
    }
    
    if(fclose(includeFile)<0) {
        perror("unable to close include file");
        exit(1);
    }
    return 1;
}

int cmd_subshell(char* args)
{
    if(strlen(args)) 
        system(args);
    else 
        system(getenv("SHELL"));
    
    printf("\n");
    return 1;
}

int cmd_set_apdufile(char* arg)
{
    REMOVE_TAILING_BLANKS(arg);
  
    if(apdu_file && apdu_file != stderr) { /* don't close stdout*/
        perror("unable to close apdu log file");      
    }
    apdu_file=NULL;
  
    if(strlen(arg)<1) {
        return 1;
    }
  
    if(!strcmp(arg,"-")) 
        apdu_file=stderr;      
    else 
        apdu_file=fopen(arg, "a");
  
    if(!apdu_file) {
        perror("unable to open apdu log file no apdu log loaded");
    } else {
        odr_setprint(print, apdu_file); 
    }
  
    return 1;
}

int cmd_set_cclfields(char* arg)
{  
#if YAZ_MODULE_ccl
    FILE *inf;

    REMOVE_TAILING_BLANKS(arg);

    bibset = ccl_qual_mk (); 
    inf = fopen (arg, "r");
    if (inf)
    {
        ccl_qual_file (bibset, inf);
        fclose (inf);
    }
#else 
    fprintf(stderr,"Not compiled with the yaz ccl module\n");
#endif
    
    return 1;
}

int cmd_set_marcdump(char* arg)
{
    if(marcdump && marcdump != stderr) { /* don't close stdout*/
        perror("unable to close apdu log file");      
    }
    marcdump=NULL;
    
    if(strlen(arg)<1) {
        return 1;
    }
  
    if(!strcmp(arg,"-")) 
        marcdump=stderr;      
    else 
        marcdump=fopen(arg, "a");
  
    if(!marcdump) {
        perror("unable to open apdu marcdump file no marcdump done\n");
    }
  
    return 1;
}

int cmd_set_proxy(char* arg) {
    if(yazProxy) free(yazProxy);
    yazProxy=NULL;

    if(strlen(arg) > 1) {
        yazProxy=strdup(arg);
    }
    return 1;
}

/* 
   this command takes 3 arge {name class oid} 
 */
int cmd_register_oid(char* args) {
    static struct {
        char* className;
        oid_class oclass;
    } oid_classes[] = {
        {"appctx",CLASS_APPCTX},
        {"absyn",CLASS_ABSYN},
        {"attset",CLASS_ATTSET},
        {"transyn",CLASS_TRANSYN},
        {"diagset",CLASS_DIAGSET},
        {"recsyn",CLASS_RECSYN},
        {"resform",CLASS_RESFORM},
        {"accform",CLASS_ACCFORM},
        {"extserv",CLASS_EXTSERV},
        {"userinfo",CLASS_USERINFO},
        {"elemspec",CLASS_ELEMSPEC},
        {"varset",CLASS_VARSET},
        {"schema",CLASS_SCHEMA},
        {"tagset",CLASS_TAGSET},
        {"general",CLASS_GENERAL},
        {0,0}
    };
    char oname_str[101], oclass_str[101], oid_str[101];  
    char* name;
    int i;
    oid_class oidclass = CLASS_GENERAL;
    int val = 0, oid[OID_SIZE];
    struct oident * new_oident=NULL;
    
    if (sscanf (args, "%100[^ ] %100[^ ] %100s",
                oname_str,oclass_str, oid_str) < 1) {
        printf("Error in regristrate command \n");
        return 0;
    }
    
    for (i = 0; oid_classes[i].className; i++) {
        if (!strcmp(oid_classes[i].className, oclass_str))
        {
            oidclass=oid_classes[i].oclass;
            break;
        }
    }
    
    if(!(oid_classes[i].className)) {
        printf("Unknonwn oid class %s\n",oclass_str);
        return 0;
    }

    i = 0;
    name = oid_str;
    val = 0;
    
    while (isdigit (*name))
    {
        val = val*10 + (*name - '0');
        name++;
        if (*name == '.')
        {
            if (i < OID_SIZE-1)
                oid[i++] = val;
            val = 0;
            name++;
        }
    }
    oid[i] = val;
    oid[i+1] = -1;
    
    new_oident=oid_addent (oid,PROTO_GENERAL,oidclass,oname_str,VAL_DYNAMIC);  
    if(strcmp(new_oident->desc,oname_str)) {
        fprintf(stderr,"oid is already named as %s, regristration faild\n",
                new_oident->desc);
    }
    return 1;  
}

int cmd_push_command(char* arg) {
#if HAVE_READLINE_HISTORY_H
    if(strlen(arg)>1) 
        add_history(arg);
#else 
    fprintf(stderr,"Not compiled with the readline/history module\n");
#endif
    return 1;
}

void source_rcfile() {
    /*  Look for a $HOME/.yazclientrc and source it if it exists */
    struct stat statbuf;
    char buffer[1000];
    char* homedir=getenv("HOME");
    if(!homedir) return;
    
    sprintf(buffer,"%s/.yazclientrc",homedir);
    
    if(stat(buffer,&statbuf)==0) {
        cmd_source(buffer);
    }
    
    if(stat(".yazclientrc",&statbuf)==0) {
        cmd_source(".yazclientrc");
    }
}


static void initialize(void)
{
#if YAZ_MODULE_ccl
    FILE *inf;
#endif
    if (!(out = odr_createmem(ODR_ENCODE)) ||
        !(in = odr_createmem(ODR_DECODE)) ||
        !(print = odr_createmem(ODR_PRINT)))
    {
        fprintf(stderr, "failed to allocate ODR streams\n");
        exit(1);
    }
    oid_init();

    setvbuf(stdout, 0, _IONBF, 0);
    if (apdu_file)
        odr_setprint(print, apdu_file);

#if YAZ_MODULE_ccl
    bibset = ccl_qual_mk (); 
    inf = fopen (ccl_fields, "r");
    if (inf)
    {
        ccl_qual_file (bibset, inf);
        fclose (inf);
    }
#endif
    cmd_base("Default");

#if HAVE_READLINE_READLINE_H
    rl_attempted_completion_function = (CPPFunction*)readline_completer;
#endif

    source_rcfile();
}


#if HAVE_GETTIMEOFDAY
struct timeval tv_start, tv_end;
#endif

void wait_and_handle_responce() 
{
    
    int res;
    char *netbuffer= 0;
    int netbufferlen = 0;
    Z_APDU *apdu;
    
    
    if (conn)
    {
        do
        {
            if ((res = cs_get(conn, &netbuffer, &netbufferlen)) < 0)
            {
                printf("Target closed connection\n");
                close_session ();
                break;
            }
            if (!res)
            {
                printf("Target closed connection.\n");
                close_session ();
                break;
            }
            odr_reset(in); /* release APDU from last round */
            record_last = 0;
            odr_setbuf(in, netbuffer, res, 0);
            if (!z_APDU(in, &apdu, 0, 0))
            {
                odr_perror(in, "Decoding incoming APDU");
                fprintf(stderr, "[Near %d]\n", odr_offset(in));
                fprintf(stderr, "Packet dump:\n---------\n");
                odr_dumpBER(stderr, netbuffer, res);
                fprintf(stderr, "---------\n");
                if (apdu_file)
                    z_APDU(print, &apdu, 0, 0);
                close_session ();
                break;
            }
            if (apdu_file && !z_APDU(print, &apdu, 0, 0))
            {
                odr_perror(print, "Failed to print incoming APDU");
                odr_reset(print);
                continue;
            }
            switch(apdu->which)
            {
            case Z_APDU_initResponse:
                process_initResponse(apdu->u.initResponse);
                break;
            case Z_APDU_searchResponse:
                process_searchResponse(apdu->u.searchResponse);
                break;
            case Z_APDU_scanResponse:
                process_scanResponse(apdu->u.scanResponse);
                break;
            case Z_APDU_presentResponse:
                print_refid (apdu->u.presentResponse->referenceId);
                setno +=
                    *apdu->u.presentResponse->numberOfRecordsReturned;
                if (apdu->u.presentResponse->records)
                    display_records(apdu->u.presentResponse->records);
                else
                    printf("No records.\n");
                printf ("nextResultSetPosition = %d\n",
                        *apdu->u.presentResponse->nextResultSetPosition);
                break;
            case Z_APDU_sortResponse:
                process_sortResponse(apdu->u.sortResponse);
                break;
            case Z_APDU_extendedServicesResponse:
                printf("Got extended services response\n");
                process_ESResponse(apdu->u.extendedServicesResponse);
                break;
            case Z_APDU_close:
                printf("Target has closed the association.\n");
                process_close(apdu->u.close);
                break;
            case Z_APDU_resourceControlRequest:
                process_resourceControlRequest
                    (apdu->u.resourceControlRequest);
                break;
            case Z_APDU_deleteResultSetResponse:
                process_deleteResultSetResponse(apdu->u.
                                                deleteResultSetResponse);
                break;
            default:
                printf("Received unknown APDU type (%d).\n", 
                       apdu->which);
                close_session ();
            }
        }
        while (conn && cs_more(conn));
#if HAVE_GETTIMEOFDAY
        gettimeofday (&tv_end, 0);
#if 0
        printf ("S/U S/U=%ld/%ld %ld/%ld",
                (long) tv_start.tv_sec,
                (long) tv_start.tv_usec,
                (long) tv_end.tv_sec,
                (long) tv_end.tv_usec);
#endif
        printf ("Elapsed: %.6f\n",
                (double) tv_end.tv_usec / 1e6 + tv_end.tv_sec -
                ((double) tv_start.tv_usec / 1e6 + tv_start.tv_sec));
#endif
    }
    xfree (netbuffer);
}


static int cmd_help (char *line);

static struct {
    char *cmd;
    int (*fun)(char *arg);
    char *ad;
    char *(*rl_completerfunction)(const char *text, int state);
    int complete_filenames;
} cmd[] = {
    {"open", cmd_open, "('tcp'|'ssl')':<host>[':'<port>][/<db>]",NULL,0},
    {"quit", cmd_quit, "",NULL,0},
    {"find", cmd_find, "<query>",NULL,0},
    {"delete", cmd_delete, "<setname>",NULL,0},
    {"base", cmd_base, "<base-name>",NULL,0},
    {"show", cmd_show, "<rec#>['+'<#recs>['+'<setname>]]",NULL,0},
    {"scan", cmd_scan, "<term>",NULL,0},
    {"sort", cmd_sort, "<sortkey> <flag> <sortkey> <flag> ...",NULL,0},
    {"sort+", cmd_sort_newset, "<sortkey> <flag> <sortkey> <flag> ...",NULL,0},
    {"authentication", cmd_authentication, "<acctstring>",NULL,0},
    {"lslb", cmd_lslb, "<largeSetLowerBound>",NULL,0},
    {"ssub", cmd_ssub, "<smallSetUpperBound>",NULL,0},
    {"mspn", cmd_mspn, "<mediumSetPresentNumber>",NULL,0},
    {"status", cmd_status, "",NULL,0},
    {"setnames", cmd_setnames, "",NULL,0},
    {"cancel", cmd_cancel, "",NULL,0},
    {"format", cmd_format, "<recordsyntax>",complete_format,0},
    {"schema", cmd_schema, "<schema>",complete_schema,0},
    {"elements", cmd_elements, "<elementSetName>",NULL,0},
    {"close", cmd_close, "",NULL,0},
    {"attributeset", cmd_attributeset, "<attrset>",complete_attributeset,0},
    {"querytype", cmd_querytype, "<type>",complete_querytype,0},
    {"refid", cmd_refid, "<id>",NULL,0},
    {"itemorder", cmd_itemorder, "ill|item <itemno>",NULL,0},
    {"update", cmd_update, "<item>",NULL,0},
    {"packagename", cmd_packagename, "<packagename>",NULL,0},
    {"proxy", cmd_proxy, "[('tcp'|'ssl')]<host>[':'<port>]",NULL,0},
    {".", cmd_source, "<filename>",NULL,1},
    {"!", cmd_subshell, "Subshell command",NULL,0},
    {"set_apdufile", cmd_set_apdufile, "<filename>",NULL,0},
    {"set_marcdump", cmd_set_marcdump," <filename>",NULL,0},
    {"set_cclfields", cmd_set_cclfields,"<filename>",NULL,1}, 
    {"register_oid",cmd_register_oid,"<name> <class> <oid>",NULL,0},
    {"push_command",cmd_push_command,"<command>",command_generator,0},
    /* Server Admin Functions */
    {"adm-reindex", cmd_adm_reindex, "<database-name>",NULL,0},
    {"adm-truncate", cmd_adm_truncate, "('database'|'index')<object-name>",NULL,0},
    {"adm-create", cmd_adm_create, "",NULL,0},
    {"adm-drop", cmd_adm_drop, "('database'|'index')<object-name>",NULL,0},
    {"adm-import", cmd_adm_import, "<record-type> <dir> <pattern>",NULL,0},
    {"adm-refresh", cmd_adm_refresh, "",NULL,0},
    {"adm-commit", cmd_adm_commit, "",NULL,0},
    {"adm-shutdown", cmd_adm_shutdown, "",NULL,0},
    {"adm-startup", cmd_adm_startup, "",NULL,0},
    {"help", cmd_help, "", NULL},
    {0,0,0,0,0}
};

static int cmd_help (char *line)
{
    int i;
    char topic[21];
    
    *topic = 0;
    sscanf (line, "%20s", topic);

    if (*topic == 0)
        printf("Commands:\n");
    for (i = 0; cmd[i].cmd; i++)
        if (*topic == 0 || strcmp (topic, cmd[i].cmd) == 0)
            printf("   %s %s\n", cmd[i].cmd, cmd[i].ad);
    if (strcmp (topic, "find") == 0)
    {
        printf ("RPN:\n");
        printf (" \"term\"                        Simple Term\n");
        printf (" @attr [attset] type=value op  Attribute\n");
        printf (" @and opl opr                  And\n");
        printf (" @or opl opr                   Or\n");
        printf (" @not opl opr                  And-Not\n");
        printf (" @set set                      Result set\n");
        printf ("\n");
        printf ("Bib-1 attribute types\n");
        printf ("1=Use:         ");
        printf ("4=Title 7=ISBN 8=ISSN 30=Date 62=Abstract 1003=Author 1016=Any\n");
        printf ("2=Relation:    ");
        printf ("1<   2<=  3=  4>=  5>  6!=  102=Relevance\n");
        printf ("3=Position:    ");
        printf ("1=First in Field  2=First in subfield  3=Any position\n");
        printf ("4=Structure:   ");
        printf ("1=Phrase  2=Word  3=Key  4=Year  5=Date  6=WordList\n");
        printf ("5=Truncation:  ");
        printf ("1=Right  2=Left  3=L&R  100=No  101=#  102=Re-1  103=Re-2\n");
        printf ("6=Completeness:");
        printf ("1=Incomplete subfield  2=Complete subfield  3=Complete field\n");
    }
    return 1;
}

void process_cmd_line(char* line)
{  
    int i,res;
    char word[32], arg[1024];
    
#if HAVE_GETTIMEOFDAY
    gettimeofday (&tv_start, 0);
#endif
    
    if ((res = sscanf(line, "%31s %1023[^;]", word, arg)) <= 0)
    {
        strcpy(word, last_cmd);
        *arg = '\0';
    }
    else if (res == 1)
        *arg = 0;
    strcpy(last_cmd, word);
    
    /* removed tailing spaces from the arg command */
    { 
        char* p;
        char* lastnonspace=NULL;
        p = arg;
        
        for(;*p; ++p) {
            if(!isspace(*p)) {
                lastnonspace = p;
            }
        }
        if(lastnonspace) 
            *(++lastnonspace) = 0;
    }
    

    for (i = 0; cmd[i].cmd; i++)
        if (!strncmp(cmd[i].cmd, word, strlen(word)))
        {
            res = (*cmd[i].fun)(arg);
            break;
        }

    if (!cmd[i].cmd) /* dump our help-screen */
    {
        printf("Unknown command: %s.\n", word);
        cmd_help ("");

    }

    if(apdu_file) fflush(apdu_file);
    
    if (res >= 2)
        wait_and_handle_responce();

    if(apdu_file)
        fflush(apdu_file);
    if(marcdump)
        fflush(marcdump);
}


char *command_generator(const char *text, int state) 
{
    static int idx; 
    if (state==0) {
        idx = 0;
    }
    for( ; cmd[idx].cmd; ++idx) {
        if (!strncmp(cmd[idx].cmd,text,strlen(text))) {
            ++idx;  /* skip this entry on the next run */
            return strdup(cmd[idx-1].cmd);
        }
    }
    return NULL;
}


/* 
   This function only known how to complete on the first word
*/
char ** readline_completer(char *text, int start, int end) {
#if HAVE_READLINE_READLINE_H

    if(start == 0) {
#ifdef RL_READLINE_VERSION
        char** res=rl_completion_matches(text,
                                      command_generator); 
#else
        char** res=completion_matches(text,
                                      (CPFunction*)command_generator); 
#endif
        rl_attempted_completion_over = 1;
        return res;
    } else {
        char arg[1024],word[32];
        int i=0 ,res;
        if ((res = sscanf(rl_line_buffer, "%31s %1023[^;]", word, arg)) <= 0) {     
            rl_attempted_completion_over = 1;
            return NULL;
        }
    
        if(start != strlen(word) +1 ) {
            rl_attempted_completion_over = 1;
            return 0;
        }
        for (i = 0; cmd[i].cmd; i++) {
            if (!strncmp(cmd[i].cmd, word, strlen(word))) {
                break;
            }
        }
    

        if(!cmd[i].complete_filenames) 
            rl_attempted_completion_over = 1;    
        if(cmd[i].rl_completerfunction) {
#ifdef RL_READLINE_VERSION
            char** res=
                rl_completion_matches(text,
                                   cmd[i].rl_completerfunction);
#else
            char** res=
                completion_matches(text,
                                   (CPFunction*)cmd[i].rl_completerfunction);
#endif
            rl_attempted_completion_over = 1;    
            return res;
        } else {
            rl_attempted_completion_over = 1;
            return 0;
        }
    }
#else 
    return 0;
#endif 
}


static void client(void)
{
    char line[1024];

    line[1023] = '\0';

#if HAVE_GETTIMEOFDAY
    gettimeofday (&tv_start, 0);
#endif

    while (1)
    {
        char *line_in = NULL;
#if HAVE_READLINE_READLINE_H
        if (isatty(0))
        {
            line_in=readline(C_PROMPT);
            if (!line_in)
                break;
#if HAVE_READLINE_HISTORY_H
            if (*line_in)
                add_history(line_in);
#endif
            strncpy(line, line_in, 1023);
            free (line_in);
        }
#endif 
        if (!line_in)
        {
            char *end_p;
            printf (C_PROMPT);
            fflush(stdout);
            if (!fgets(line, 1023, stdin))
                break;
            if ((end_p = strchr (line, '\n')))
                *end_p = '\0';
        }
        process_cmd_line(line);
    }
}

int main(int argc, char **argv)
{
    char *prog = *argv;
    char *open_command = 0;
    char *auth_command = 0;
    char *arg;
    int ret;
    
    while ((ret = options("k:c:a:m:v:p:u:", argv, argc, &arg)) != -2)
    {
        switch (ret)
        {
        case 0:
            if (!open_command)
            {
                open_command = (char *) xmalloc (strlen(arg)+6);
                strcpy (open_command, "open ");
                strcat (open_command, arg);
            }
            break;
        case 'k':
            kilobytes = atoi(arg);
            break;
        case 'm':
            if (!(marcdump = fopen (arg, "a")))
            {
                perror (arg);
                exit (1);
            }
            break;
    case 'c':
        strncpy (ccl_fields, arg, sizeof(ccl_fields)-1);
        ccl_fields[sizeof(ccl_fields)-1] = '\0';
        break;
        case 'a':
            if (!strcmp(arg, "-"))
                apdu_file=stderr;
            else
                apdu_file=fopen(arg, "a");
            break;
    case 'p':
        yazProxy=strdup(arg);
        break;
        case 'u':
            if (!auth_command)
            {
                auth_command = (char *) xmalloc (strlen(arg)+6);
                strcpy (auth_command, "auth ");
                strcat (auth_command, arg);
            }
            break;
        case 'v':
            yaz_log_init (yaz_log_mask_str(arg), "", NULL);
            break;
        default:
            fprintf (stderr, "Usage: %s [-m <marclog>] [ -a <apdulog>] "
                     "[-c cclfields]\n      [-p <proxy-addr>] [-u <auth>] "
                     "[-k size] [<server-addr>]\n",
                     prog);
            exit (1);
        }      
    }
    initialize();
    if (auth_command)
    {
#ifdef HAVE_GETTIMEOFDAY
        gettimeofday (&tv_start, 0);
#endif
        process_cmd_line (auth_command);
#if HAVE_READLINE_HISTORY_H
        add_history(auth_command);
#endif
        xfree(auth_command);
    }
    if (open_command)
    {
#ifdef HAVE_GETTIMEOFDAY
        gettimeofday (&tv_start, 0);
#endif
        process_cmd_line (open_command);
#if HAVE_READLINE_HISTORY_H
        add_history(open_command);
#endif
        xfree(open_command);
    }
    client ();
    exit (0);
}


/*
 * Local variables:
 * tab-width: 4
 * c-basic-offset: 4
 * End:
 */
